#include "m_pd.h"
#include "math.h"
#define SGN(x) ( ((x) < 0) ? -1 : ( ((x) == 0 ) ? 0 : 1) )

// MY DECLARATION ---------------------------------------
// WDF BOX ----------------------------------------------
struct _WDF {
	double Rp, DW, RW;
};

double voltage_WDF (struct _WDF *WDF) {
	return WDF->DW + WDF->RW;
}

double current_WDF (struct _WDF *WDF) {
	return (WDF->DW-WDF->RW)/WDF->Rp;
}

// RES BOX ----------------------------------------------
struct _RES {
	double Rp, DW, RW;
};

void init_RES(struct _RES *RES, double R) {
	RES->Rp = R;
	RES->DW = 0;
	RES->RW = 0;
}

void compute_RES(struct _RES *RES){
	RES->RW = 0;
}

void reset_RES(struct _RES *RES) {
	RES->DW = RES->RW = 0;
}


//CAP BOX ----------------------------------------------
struct _CAP {
	double Rp, DW, RW;
	double s;
};

void init_CAP(struct _CAP *CAP, double C) {
	CAP->Rp = 1/(88200*C);
	CAP->DW = 0;
	CAP->RW = 0;
	CAP->s = 0;
}

void compute_CAP(struct _CAP *CAP) {
	CAP->RW = CAP->s;
	CAP->s = CAP->DW;
}

void reset_CAP(struct _CAP *CAP) {
	CAP->DW = CAP->RW = CAP->s = 0;
}

// IND BOX ----------------------------------------------
struct _IND {
	double Rp, DW, RW;
	double s;
};

void init_IND(struct _IND *IND, double L) {
	IND->Rp = 88200*L;
	IND->DW = 0;
	IND->RW = 0;
	IND->s = 0;
}

void compute_IND(struct _IND *IND) {
	IND->RW = -IND->s;
	IND->s = IND->DW;
}

void reset_IND(struct _IND *IND) {
	IND->DW = IND->RW = IND->s = 0;
}

// GEN BOX ----------------------------------------------
struct _GEN {
	double Rp, DW, RW;
	double E;
};

void init_GEN(struct _GEN *GEN, double E, double Ri) {
	GEN->Rp = Ri;
	GEN->DW = 0;
	GEN->RW = E/2;
	GEN->E = E;
}

void compute_GEN (struct _GEN *GEN) {
	GEN->RW = GEN->E/2;
}

void reset_GEN(struct _GEN *GEN) {
	GEN->DW = 0;
}

// BILINEAR DIODE BOX -----------------------------------

struct _DD {
	double Rp, DW, RW;
	double Rpp, Rpn;
};

void init_DD(struct _DD *DD, double Rpp, double Rpn){
	DD->Rp = Rpp;
	DD->DW = 0;
	DD->RW = 0;
	DD->Rpp = Rpp;
	DD->Rpn = Rpn;
}

void compute_DD(struct _DD *DD) {
	DD->RW = 0;
	if (DD->DW+DD->RW>0) DD->Rp = DD->Rpp;
	else DD->Rp = DD->Rpn;
}

void reset_DD(struct _DD *DD) {
	DD->DW = DD->RW = 0;
}

// CHILD-LANGMUIR DIODE BOX -----------------------------
struct _D_CL {
	double Rp, DW, RW;
	double kD, uD, u, i;
};

void init_D_CL(struct _D_CL *D_CL, double kD, double uD) {
	D_CL->Rp = 1;
	D_CL->DW = 0;
	D_CL->RW = 0;
	D_CL->kD = kD;
	D_CL->uD = uD;
	D_CL-> i = -60;
	D_CL-> u = 10;
}

void compute_D_CL(struct _D_CL *D_CL) {
	D_CL->RW = 0;
	if ((D_CL->DW+D_CL->RW)<1e-60) {
		D_CL->Rp = (1/D_CL->kD)*pow(10,-60*(1-D_CL->uD));
	}
	else {
		D_CL->Rp = (1/D_CL->kD)*pow(D_CL->DW+D_CL->RW,(1-D_CL->uD));
	}
}

void reset_D_CL(struct _D_CL *D_CL) {
	D_CL->DW = D_CL->RW = 0;
	D_CL->Rp = 1;
}

// TRANSFORMER BOX --------------------------------------
struct _tr {
	double Rp, DW, RW;
	double N;
	// pointer to struct on port
	struct _WDF *port;
};

void init_tr (struct _tr *tr, double N, struct _WDF *port) {
	tr->N = N;
	tr->port = port;
	tr->Rp = tr->port->Rp/(N*N);
	tr->DW = 0;
	tr->RW = 0;
}

void compute_tr (struct _tr *tr, int direction) {
	if (direction == 0) {  /* leaves -> root */
		tr->Rp = tr->port->Rp/(tr->N*tr->N);
		tr->RW = tr->port->RW/(tr->N);
	}
	else { /* root -> leaves*/
		tr->port->DW = tr->DW*tr->N;
	}
}

void reset_tr(struct _tr *tr) {
	tr->DW = tr->RW = 0;
}

// PARALLEL ADAPTOR BOX ---------------------------------
struct _pa {
	double Rp, DW, RW;
	double g; // temporary
	// pointers to struct on ports
	struct _WDF *port1, *port2;
};

void init_pa (struct _pa *pa, struct _WDF *port1, struct _WDF *port2) {
	pa->port1 = port1;
	pa->port2 = port2;
	pa->Rp = (pa->port1->Rp*pa->port2->Rp)/
	(pa->port1->Rp+pa->port2->Rp);
	pa->g = pa->Rp;
}

void compute_pa (struct _pa *pa, int direction) {
	if (direction == 0) /* leaves -> root */ {
		pa->Rp = (pa->port1->Rp*pa->port2->Rp)/(pa->port1->Rp+pa->port2->Rp);
		pa->g = pa->Rp;
		pa->RW = (pa->g/pa->port1->Rp)*pa->port1->RW+(pa->g/pa->port2->Rp)*pa->port2->RW;
	}
	else /* root -> leaves */ {
		pa->g = pa->Rp;
		pa->port1->DW = pa->RW+pa->DW-pa->port1->RW;
		pa->port2->DW = pa->RW+pa->DW-pa->port2->RW;
	}
}

void reset_pa(struct _pa *pa) {
	pa->DW = pa->RW = 0;
	pa->Rp = (pa->port1->Rp*pa->port2->Rp)/(pa->port1->Rp+pa->port2->Rp);
	pa->g = pa->Rp;
}

// SERIAL ADAPTOR BOX -----------------------------------
struct _sa {
	double Rp, DW, RW;
	double g; // temporary
	// pointers to struct on ports
	struct _WDF *port1, *port2;
};

void init_sa (struct _sa *sa, struct _WDF *port1, struct _WDF *port2) {
	sa->port1 = port1;
	sa->port2 = port2;
	sa->Rp = sa->port1->Rp+sa->port2->Rp;
	sa->g = 1/sa->Rp;
}

void compute_sa (struct _sa *sa, int direction) {
	if (direction == 0) { /* leaves -> root */
		sa->Rp = sa->port1->Rp+sa->port2->Rp;
		sa->g = 1/sa->Rp;
		sa->RW = -sa->port1->RW-sa->port2->RW;
	}
	else { /* root -> leaves */
		sa->g = 1/sa->Rp;
		sa->port1->DW = sa->port1->RW-(sa->port1->Rp*sa->g*(sa->DW-sa->RW));
		sa->port2->DW = sa->port2->RW-(sa->port2->Rp*sa->g*(sa->DW-sa->RW));
	}
}

void reset_sa(struct _sa *sa) {
	sa->DW = sa->RW = 0;
	sa->Rp = sa->port1->Rp+sa->port2->Rp;
	sa->g = 1/sa->Rp;
}

// various ----------------------------------------------
double read_DW (struct _WDF *p) {
	return p->DW;
}
void write_RW (struct _WDF *p, double c) {
	p->RW = c;
}


static t_class *triode_tilde_class;

typedef struct _triode_tilde {
	t_object x_obj;
	t_sample f;

	t_sample Rin;
	double Vpk,Vgk,E1,Ipk,RNL;
	t_outlet *my_out;

	struct _RES R0;
	struct _CAP C0;
	struct _GEN V_Rp;
	struct _CAP Ck;
	struct _RES Rk;


	struct _D_CL D;

	struct _RES Rg;
	struct _CAP Ci;
	struct _RES Vin;
	struct _RES Ri;

	struct _tr T;
	struct _sa s5;
	struct _pa p4;
	struct _sa s4;
	struct _sa s2;
	struct _pa p3;
	struct _pa p1;
	struct _sa s3;
	struct _pa p2;
	struct _sa s1;
} t_triode_tilde;

void triode_tilde_bang(t_triode_tilde *x) {
// Bang to reset WDF
	x->Vpk = 1;
	x->Vgk = 1;
	x-> E1 = 0;
	x->Ipk = 0;
	x->RNL = 0;

	reset_RES(&x->R0);
	reset_CAP(&x->C0);
	reset_GEN(&x->V_Rp);
	reset_CAP(&x->Ck);
	reset_RES(&x->Rk);
	reset_D_CL(&x->D);

	reset_RES(&x->Rg);

	reset_CAP(&x->Ci);
	reset_RES(&x->Vin);
	reset_RES(&x->Ri);
	reset_tr(&x->T);
	reset_sa(&x->s5);
	reset_pa(&x->p4);
	reset_sa(&x->s4);
	reset_sa(&x->s2);
	reset_pa(&x->p3);
	reset_pa(&x->p1);
	reset_sa(&x->s3);
	reset_pa(&x->p2);
	reset_sa(&x->s1);
}

t_int *triode_tilde_perform(t_int *w) {
	t_triode_tilde *x = (t_triode_tilde *)(w[1]);
	t_sample *in = (t_sample *)(w[2]);
	t_sample *out = (t_sample *)(w[3]);
	t_sample *my_out = (t_sample *)(w[4]);
	int n = (int)(w[5]);

	t_sample my_in,my_out_1,my_out_2;
	x->Vin.Rp = x->Rin;

	while (n--) {
		write_RW((struct _WDF *) &x->Vin,(double)-*in++/2);
		compute_CAP(&x->Ci);
		compute_D_CL(&x->D);
		compute_CAP(&x->Ck);
		compute_CAP(&x->C0);
		compute_tr(&x->T,0);
		compute_sa(&x->s5,0);
		compute_pa(&x->p4,0);
		compute_sa(&x->s4,0);
		compute_sa(&x->s2,0);
		compute_pa(&x->p3,0);
		compute_pa(&x->p1,0);
		compute_sa(&x->s3,0);
		compute_pa(&x->p2,0);
		compute_sa(&x->s1,0);
		x->Vgk = voltage_WDF((struct _WDF *) &x->D);
		x->Vpk = -voltage_WDF((struct _WDF *) &x->s1);
		x->E1 = (x->Vpk/600)*log(1+exp(600*((1/100)+ (x->Vgk/sqrt(300+(x->Vpk*x->Vpk))))));
		x->Ipk = ((pow(x->E1,1.4))/1060)*(1+SGN(x->E1));
		x->RNL = x->Vpk/x->Ipk;
		x->s1.DW = x->s1.RW*(x->RNL-x->s1.Rp)/(x->RNL+x->s1.Rp);
		*out++ = (t_sample) (-10000*voltage_WDF((struct _WDF *) &x->R0));
		*my_out++ = x->Vgk;
		compute_sa(&x->s1,1);
		compute_pa(&x->p2,1);
		compute_sa(&x->s3,1);
		compute_pa(&x->p1,1);
		compute_pa(&x->p3,1);
		compute_sa(&x->s2,1);
		compute_sa(&x->s4,1);
		compute_pa(&x->p4,1);
		compute_sa(&x->s5,1);
		compute_tr(&x->T,1);
	}
	return (w+6);
}


void triode_tilde_dsp(t_triode_tilde *x, t_signal **sp) {
	dsp_add(triode_tilde_perform, 5, x, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec, sp[0]->s_n);
}


void *triode_tilde_new(t_floatarg f) {
	t_triode_tilde *x = (t_triode_tilde *)pd_new(triode_tilde_class);


	post("WDF Triode (Child-Langmuir diode model + Resistive Load)");
	post("D. Perla, L. Gabrielli, (c) 2012, implemented after ");
	post("Pakarinen-Karjalainen IEEE ASLP 2010 paper");
	if ( f > 0 )
		x->Rin = f;
	else
		x->Rin = 50;
	x->Vpk = 1;
	x->Vgk = 1;
	x-> E1 = 0;
	x->Ipk = 0;
	x->RNL = 0;
	
	post("Rin = %g", x->Rin);

	init_RES(&x->R0,8);
	init_CAP(&x->C0,10e-9);
	init_GEN(&x->V_Rp,250,100e3);
	init_CAP(&x->Ck,10e-6);
	init_RES(&x->Rk,1e3);
	init_D_CL(&x->D,.005,2.1);
	init_RES(&x->Rg,20e3);
	init_CAP(&x->Ci,100e-9);
	init_RES(&x->Vin,50);
	init_RES(&x->Ri,1e6);
	init_tr(&x->T,100,(struct _WDF *) &x->R0);

	init_sa(&x->s5,(struct _WDF *) &x->Vin,(struct _WDF *) &x->Ci);
	init_pa(&x->p4,(struct _WDF *) &x->Ri,(struct _WDF *) &x->s5);
	init_sa(&x->s4,(struct _WDF *) &x->Rg,(struct _WDF *) &x->p4);
	init_sa(&x->s2,(struct _WDF *) &x->s4,(struct _WDF *) &x->D);
	init_pa(&x->p3,(struct _WDF *) &x->Rk,(struct _WDF *) &x->Ck);

	init_pa(&x->p1,(struct _WDF *) &x->s2,(struct _WDF *) &x->p3);
	init_sa(&x->s3,(struct _WDF *) &x->C0,(struct _WDF *) &x->T);
	init_pa(&x->p2,(struct _WDF *) &x->s3,(struct _WDF *) &x->V_Rp);
	init_sa(&x->s1,(struct _WDF *) &x->p1,(struct _WDF *) &x->p2);

	floatinlet_new(&x->x_obj, &x->Rin);
	outlet_new(&x->x_obj, &s_signal);
	x->my_out = outlet_new(&x->x_obj, &s_signal);
	return (void *)x;
}

void triode_tilde_setup(void) {
	triode_tilde_class = class_new(gensym("triode~"),(t_newmethod)triode_tilde_new,0, sizeof(t_triode_tilde),CLASS_DEFAULT, A_DEFFLOAT, 0);

	class_addbang(triode_tilde_class,triode_tilde_bang);
	class_addmethod(triode_tilde_class,(t_method)triode_tilde_dsp, gensym("dsp"), 0);
	CLASS_MAINSIGNALIN(triode_tilde_class, t_triode_tilde, f);

}
